/* Code written by Per Winkvist (Per.Winkvist@UK.COM)
 * In 1999
 * See README and COPYING for further details
 *
 */


#include "lilo_data.h"
#include "memwatch.h"
#include <string.h>
#include <signal.h>

int strcasecmp (const char *s1, const char *s2);

int count_common (LLIST *lilo_data, char *word);
int count_in_common (LLIST *lilo_data, int start_node);


void free_lilo_data_one (LILO_ONE_OPTION *ptr)
{
  if (ptr == NULL)
    return;
  
  free (ptr->data_name);
  
  if (ptr->data_args != NULL)
    free (ptr->data_args);

  if (ptr->data_pre_req != NULL)
    free_option_pre_req(ptr->data_pre_req);

  if (ptr->data_post_req != NULL)
    free_option_post_req(ptr->data_post_req);

  if (ptr->carg_words != NULL)
    free_option_carg(ptr->carg_words);

  free (ptr);
}



void free_lilo_data(LLIST *lilo_data)
{
  int i,size;
  LILO_ONE_OPTION *ptr;

  size = llist_get_size(lilo_data);

  for (i=0;i<size;i++) {
    ptr = llist_pop_min(lilo_data);
    
    free (ptr->data_name);
    if (ptr->data_args != NULL)
      free (ptr->data_args);

    if (ptr->data_pre_req != NULL)
      free_option_pre_req(ptr->data_pre_req);

    if (ptr->data_post_req != NULL)
      free_option_post_req(ptr->data_post_req);

    if (ptr->carg_words != NULL)
      free_option_carg(ptr->carg_words);

    free (ptr);
  }

  llist_free (lilo_data);
}

int get_common_node (LLIST *lilo_data, char *section_name, int image)
{
  LILO_ONE_OPTION *ptr;
  int count=-1,size,i;

  size = llist_get_size (lilo_data);
  if (size > 0 && (strcasecmp(section_name,"global") == 0))
    return 0; /* The global section always starts on node 0 */

  for (i=0;i<size;i++) {
    ptr = llist_get_node (lilo_data,i);
    if (strcasecmp (ptr->data_name,section_name) == 0) {
      count++;
      if (count == image)
        return i;
    }
  }
  printf("The %s you tried to get was out of reach!\n",section_name);
  printf("Tried to get %d, there is just %d %ss\n",image,
         count_common(lilo_data,section_name),section_name);
  raise(SIGSEGV);
  exit (EXIT_FAILURE);
}


int get_image_node (LLIST *lilo_data, int image)
{
  return get_common_node (lilo_data,"image",image);
}


int get_other_node (LLIST *lilo_data, int other)
{
  return get_common_node (lilo_data,"other",other);
}


int count_common (LLIST *lilo_data, char *section_name)
{
  LILO_ONE_OPTION *ptr;
  int i,count=0,size;

  size = llist_get_size(lilo_data);
  
  for (i=0;i<size;i++) {
    ptr = llist_get_node(lilo_data,i);
    if (strcasecmp(ptr->data_name,section_name) == 0)
      count++;
  }
  return count;
}


int count_images (LLIST *lilo_data)
{
  return count_common (lilo_data,"image");
}


int count_others (LLIST *lilo_data)
{
  return count_common (lilo_data,"other");
}


/*
 * This function returns the amount of elements in a section
 * NOT INCLUDING the section name itself (image/other)
 *
 * DO NOT USE THIS ONE DIRECTLY TO COUNT IN GLOBAL!!!!
 * We don't know what is on the first node !!!!
 */
int count_in_common (LLIST *lilo_data, int start_node)
{
  LILO_ONE_OPTION *ptr;
  int i,count=0,size;

  size = llist_get_size(lilo_data);

  /* If we are counting in image/other, set startnode
   * to second node in the list then... we don't want to
   * start by counting the word image or other.
   */
  if (start_node + 1 < size) {
    ptr = llist_get_node(lilo_data,start_node);
  
    if (strcasecmp(ptr->data_name,"image") == 0 ||
        strcasecmp(ptr->data_name,"other") == 0)
      start_node ++;
    /* done increasing start_node */
  }

  for (i=start_node;i<size;i++) {
    count++; /* Inc the statements-in-this-image counter */
    ptr = llist_get_node(lilo_data,i);
    if (strcasecmp(ptr->data_name,"image") == 0 ||
        strcasecmp(ptr->data_name,"other") == 0)
      return --count; /* We have found a new image/other, lets return */
    /* We use "--count" since we don't wanna count the next image/other */
  }
  /*  printf("The size of the list is :%d\n",llist_get_size(lilo_data));
   printf("I found %d things when starting at node %d\n",count,start_node);
   */
  return count; /* The last one lasted to the end of the file */
}


int count_in_global (LLIST *lilo_data)
{
  LILO_ONE_OPTION *ptr;

  if (llist_get_size(lilo_data) <= 0)
    return 0;

  ptr = llist_get_node(lilo_data,0); /* look at the first node */

  if (strcasecmp(ptr->data_name,"image") == 0 ||
      strcasecmp(ptr->data_name,"other") == 0)
    return 0; /* If first node is != global, then return 0 for itemcount */
  else /* else return items in the global section */
    return count_in_common (lilo_data,0); 
}

int count_in_image (LLIST *lilo_data, int i)
{
  return count_in_common (lilo_data,
                          get_image_node(lilo_data,i));
}

int count_in_other (LLIST *lilo_data, int i)
{
  return count_in_common (lilo_data,
                          get_other_node(lilo_data,i));
}

LLIST *merge_data (LLIST *src, LLIST *data)
{
  return llist_merge_lists (src,data);
}

LLIST *new_data (char *keyword, char *arg)
{
  LLIST *new_list;
  LILO_ONE_OPTION *new_data;
  char *new_key=NULL,*new_arg=NULL;
  
  new_list = llist_new ();

  new_data = malloc (sizeof (LILO_ONE_OPTION));

  if (keyword != NULL)
    new_key = malloc (strlen(keyword)+1);
  if (arg != NULL)
    new_arg = malloc (strlen(arg)+1);
  
  if (new_data  == NULL || /* Did the alloc fail ? */
      (new_key == NULL && keyword != NULL) || /* Was keyword null */
      (new_arg == NULL && arg != NULL)) { /* or did it fail ? */
    printf ("Couldn't get mem in %s at line %d!\n",__FILE__,__LINE__);
    exit(EXIT_FAILURE);
  }

  if (keyword != NULL) {
    strcpy(new_key,keyword);
    new_data->data_name = new_key;
  }
  else
    new_data->data_name = NULL;
    
  if (arg != NULL) {
    strcpy (new_arg,arg);
    new_data->data_args = new_arg;
  }
  else
    new_data->data_args = NULL;

  new_data->data_pre_req = NULL;  /* Filled in later */
  new_data->data_post_req = NULL;
  new_data->carg_words = NULL;
  
  llist_insert(new_list,new_data);

  return new_list;
}

/* This function adds arg to the args string and returns it */
char *merge_args (char *args, char *arg)
{
  char *new_arg;

  if (args == NULL) /* Is arg empty ? */
    return arg;
  if (arg == NULL) /* Is args empty ? */
    return args;
  if (args == arg) /* Are they the same ? */
    return args;
  
  if ((new_arg = calloc(strlen(args)+strlen(arg)+1,sizeof(char))) == NULL) {
    printf ("Couldn't get mem in %s at line %d!\n",__FILE__,__LINE__);
    exit(EXIT_FAILURE);
  }

  strcpy (new_arg,args);
  strcat (new_arg,arg);
  free(arg);
  free(args);

  return new_arg;
}

void internal_print_data (LILO_ONE_OPTION *data)
{
  static int done_glob = 0;

  if ((strncmp(data->data_name,"image",5) == 0) ||
      (strncmp(data->data_name,"other",5) == 0)) {
    printf("\n%s = \"%s\"\n",data->data_name,data->data_args);
    done_glob = 1; /* Printed image/other and done w/ globals */
  } else {
    if (done_glob != 0)
      printf("\t"); /* Lets indent if no more globals */

    printf("%s",data->data_name);
    if (data->data_args != NULL) /* do we have an argument ? */
      printf(" = %s\n",data->data_args);
    else
      printf("\n");
  }
}


void print_data (LLIST *lilo_data)
{
  if (lilo_data == NULL) {
    printf("The lilo_data list was NULL!\nExiting...\n");
    printf("FILE : %s, ROW : %d\n",__FILE__,__LINE__);
    exit(EXIT_FAILURE);
  }

  printf("There are %d elements in the list\n",llist_get_size(lilo_data));
  llist_for_each (lilo_data,(void *) internal_print_data);
}

void fill_data_with_reqs(LLIST *lilo_data)
{
  LLIST *lilo_ops;
  LILO_ONE_OPTION *ptr;
  int i,max;

  if (lilo_data == NULL)
    return ; /* if the list is empty, just return */
    
  lilo_ops = llist_new(); /* Init the linked list */
  load_options (lilo_ops,"lilo_options.dat");

  max = llist_get_size(lilo_data);
  for (i=0;i<max;i++) {
    ptr = llist_get_node(lilo_data,i);
    ptr->data_pre_req = get_option_pre_req(lilo_ops,
                                           ptr->data_name);
    ptr->data_post_req = get_option_post_req(lilo_ops,
                                             ptr->data_name);
    ptr->data_argtype = get_option_argtype(lilo_ops,
                                           ptr->data_name,1);
    if (ptr->data_argtype == ARG_COMBO)
      ptr->carg_words = get_option_carg(lilo_ops,
                                        ptr->data_name,1);
    else
      ptr->carg_words = NULL;

  }
  free_options(lilo_ops); /* close the option list */
}


void fill_one_data_with_reqs(LILO_ONE_OPTION *ptr)
{
  LLIST *lilo_ops;
  
  if (ptr == NULL)
    return ; /* if the list is empty, just return */
    
  lilo_ops = llist_new(); /* Init the linked list */
  load_options (lilo_ops,"lilo_options.dat");

  
  ptr->data_pre_req = get_option_pre_req(lilo_ops,
                                           ptr->data_name);
  ptr->data_post_req = get_option_post_req(lilo_ops,
                                             ptr->data_name);
  ptr->data_argtype = get_option_argtype(lilo_ops,
                                           ptr->data_name,1);
  if (ptr->data_argtype == ARG_COMBO)
    ptr->carg_words = get_option_carg(lilo_ops,
                                        ptr->data_name,1);
  else
    ptr->carg_words = NULL;


  free_options(lilo_ops); /* close the option list */
}

/* eof */
