/* Code written by Per Winkvist (Per.Winkvist@UK.COM)
 * In 1999
 * See README and COPYING for further details
 *
 */


#include <stdlib.h>
#include <string.h>
#include "lilo_keywords.h"
#include "lilo_options.h"
#include "lilo_data.h"
#include "lilo_common.h"

#include "memwatch.h"

int strcasecmp (const char *s1, const char *s2);
void print_ids(LLIST *section);

/*
 * Function to check that the keyword 'name' is not a child
 * The only pre-reqs it can have is either "image" or "other"
 */
int keyword_check_not_child(LLIST *options, char *name)
{
  int *value;
  int i, max;
  int ret_val = 1;
  LLIST *ptr;
  int keyid_image,keyid_other;

  ptr = get_option_pre_req (options,name);
  max = llist_get_size(ptr);

  /*
   * Get ID for "image" and "other"
   * Are used when we check for pre-req values.
   * These are the only ones we accept since the
   * rest of the values should be childs for keywords
   */
  keyid_image = get_option_id(options,"image");
  keyid_other = get_option_id(options,"other");

  /*printf("__Word is %s\n",name); */
  
  for (i=0;i<max;i++) {
    value = llist_get_node (ptr,i);
    if (*value != keyid_image &&
        *value != keyid_other)
        ret_val = 0;
  }

  free_option_pre_req (ptr);
  return ret_val; /* It was okay. 'name' is not a child */
}



/*
 * Function to get all childs that belong to a keyword
 */
LLIST *get_childs_for_keyword(LLIST *lops, char *name, int sec_id)
{
  LLIST *child_list=NULL;
  LLIST *id_ptr=NULL;
  POSSIBLE_KEYWORDS *ptr;
  LILO_OPTION *lop=NULL;
  char *new_word=NULL;
  int *value=NULL;
  int parent_id;
  int i,max;
  int i2,max2;
  
  parent_id = get_option_id(lops,name);
  max = llist_get_size (lops);

  child_list = llist_new ();
  
  for (i=0;i<max;i++) {
    lop = llist_get_node (lops,i); /* Get option 'i' */
    id_ptr = get_option_pre_req (lops,lop->op_name); /* Get i's pre-req's */

    /* Go through each pre req and look if it belongs to 'name' */
    max2 = llist_get_size (id_ptr);
    for (i2=0;i2<max2;i2++) {
      value = llist_get_node(id_ptr,i2);
      if (*value == parent_id &&
          IS_OPTION_TYPE_FOR_SECTION (lop->op_type,sec_id)) {

        /* Allocate a new child entry */
        ptr = malloc (sizeof (POSSIBLE_KEYWORDS));

        /* Allocate space for the entry name */
        if ((new_word = calloc (strlen(lop->op_name)+1,sizeof(char))) == NULL) {
          perror("Couldn't calloc memory : ");
          printf("In %s at line %d\n",__FILE__,__LINE__);
          exit (EXIT_FAILURE);
        }

        /* Initialize the properties of the entry */
        strcpy (new_word,lop->op_name);
        ptr->name = new_word;

        /* Recurse and get childs for this child */
        ptr->childs = get_childs_for_keyword(lops,ptr->name,sec_id);
        if (lop->op_post) /* Is the word supposed to be "grouped" with childs ? */
          ptr->group = 1; /* Yes, group. Example "MAP-DRIVE & TO" */
        else
          ptr->group = 0; /* no */

        llist_insert (child_list,ptr);
        i2 = max2; /* Leave i2 loop */
      }
    }
    free_option_pre_req (id_ptr);
  }
  
  if (llist_get_size(child_list) <= 0) {
    llist_free (child_list);
    child_list = NULL;
  }
  return child_list;
}

/*
 * Returns how many times the 'id' has been used in
 * the current section.
 */
int keyword_count_usage (LLIST *section,int id)
{
  int i,max;
  int count=0;
  int *value;

  max = llist_get_size (section);
  
  for (i=0;i<max;i++) {
    value = llist_get_node (section,i);
    if (*value == id)
      count++;
  }
  return count;
}

/*
 * Looks if the word is a post req.
 * It does so by finding word's parent, then going through
 * parents post_reqs and comparing them with actual word
 */
int check_if_word_is_post (LLIST *lops,char *name)
{
  int parent_id=-1,child_id;
  int found=0;
  int i,max;
  int *value;
  LLIST *reqs;
  LILO_OPTION *ptr;


  ptr = get_option (lops,name); /* Get names block */
  child_id = ptr->op_id;        /* Get the id for name */

  reqs = get_option_pre_req (lops,ptr->op_name);
  max = llist_get_size (reqs);
  if (max > 0) {
    /* We found a parent */
    value = llist_get_node(reqs,0); /* Grab the first one */
    parent_id = *value; /* Set the value of the parent */
  }

  free_option_pre_req(reqs); /* Free the pre reqs */
  
  if (max <= 0)
    return 0; /* Return if no parent was found */

  ptr = get_option_by_id (lops,parent_id); /* Get parents block */

  
  reqs = get_option_post_req (lops,ptr->op_name);
  max = llist_get_size(reqs); /* Look if parent has a post-req of *name */
  
  for (i=0;i<max && !found;i++) {
    value = llist_get_node (reqs,i);
    if (*value == child_id) /* Parents postreq matches the *name */
      found = 1;
  }

  free_option_post_req (reqs); /* free the post reqs */
  return found;
}

 

/*
 * Function check if one keyword can be enabled
 */
void keyword_check_if_enabled (LLIST *lops,
                               LLIST *section_glob,
                               LLIST *section_curr,
                               LLIST *keylist,
                               int sec_id)
{
  POSSIBLE_KEYWORDS *ptr;
  LILO_OPTION *lop;
  int i;
  int max;
  int count;

  max = llist_get_size (keylist);

  for (i=0;i<max;i++) {
    ptr = llist_get_node (keylist,i);

    if (check_pre_option(lops,section_glob,ptr->name) == 1 ||
        check_pre_option(lops,section_curr,ptr->name) == 1 ||
        check_if_word_is_post(lops,ptr->name) == 1)
      ptr->enabled = KEYWORD_ENABLED;
    else
      ptr->enabled = KEYWORD_DISABLED;

    lop = get_option (lops,ptr->name);
    /* Then do use_once/use_many checking */
    if (lop->op_count > 0) { /* The keyword is limited to op_count times */
      if (sec_id == LILO_GLOBAL)
        count = keyword_count_usage(section_glob,lop->op_id);
      else
        count = keyword_count_usage(section_curr,lop->op_id);
      if (count >= lop->op_count)
        ptr->enabled = KEYWORD_USED; /* Disable keyword since it has been used */
    }

    if (ptr->childs) /* Recurse check children */
      keyword_check_if_enabled (lops,section_glob,
                                section_curr,ptr->childs,sec_id);
  }
}



/*
 * Function that will return a llist with entries of type
 * POSSIBLE_KEYWORDS for a specific section.
 */
LLIST *get_keywords_for(LLIST *lilo_data, char *sec_name, int section)
{
  POSSIBLE_KEYWORDS *ptr;
  LILO_ONE_OPTION *data;      /* Pointer to one lilo_data */
  LILO_OPTION *lop;           /* Pointer to within 'lops' */
  LLIST *lops=NULL;           /* Holds the lilo_options */
  LLIST *ret_list=NULL;       /* The *char list we'll return */
  LLIST *section_glob=NULL;   /* lilo_data from global section */
  LLIST *section_curr=NULL;   /* Holds lilo_data from current section */
  char *new_word;
  int i,sec_node=0,*temp;
  int sec_count=0,glob_count=0;
  int max_ops;
  int sec_id;                 /* LILO_{GLOBAL,IMAGE,OTHER} */

  
  if (strcasecmp(sec_name,"global") == 0)
    sec_id = LILO_GLOBAL;
  else if (strcasecmp(sec_name,"image") == 0)
    sec_id = LILO_IMAGE;
  else
    sec_id = LILO_OTHER;

  if (llist_get_size(lilo_data) > 0) {
    sec_node = get_common_node(lilo_data,sec_name,section);
    if (sec_id != LILO_GLOBAL)
      sec_count = count_in_common(lilo_data,sec_node);

    glob_count = count_in_global(lilo_data);
  }
  else
    glob_count = 0;

  /* Load the list with all options */
  lops = llist_new();
  load_options (lops,"lilo_options.dat");
  max_ops = llist_get_size(lops);

  /* First we wanna copy global section to a own list */
  section_glob = llist_new();
  for (i=0;i<glob_count;i++) {
    data = llist_get_node(lilo_data,i);
    lop = get_option(lops,data->data_name);
    temp = malloc (sizeof(int));
    *temp = lop->op_id;
    llist_insert (section_glob,temp);
  }

  /* If there are no global options, then delete the list */
  if (llist_get_size(section_glob) < 1) {
    llist_free(section_glob);
    section_glob = NULL;
  }

  /* If section passed to function is the global, then we
   * don't care to get the data for it since its the same
   * data as the one above
   */
  if (strcmp(sec_name,"global") != 0) {
    /* Then we wanna copy the current section to a own list */
    section_curr = llist_new();
    for (i=0;i<sec_count+1;i++) {
      data = llist_get_node(lilo_data,i + sec_node);
      lop = get_option(lops,data->data_name);
      temp = malloc (sizeof(int));
      *temp = lop->op_id;

      llist_insert (section_curr,temp);
    }

    /* And delete it too if it is empty */
    if (llist_get_size(section_curr) < 1) {
      llist_free(section_curr);
      section_curr = NULL;
    }
  }

  ret_list = llist_new();

  /* Now traverse through lilo options and try 'n match with
   * global options and current section.
   * If one (option in global/section match lilo_options) then
   * add it to the LLIST that we'll return
   */


  for (i=0;i<max_ops;i++) {
    lop = llist_get_node(lops,i);

/*    printf("%d. Checking for %s [type == %d]\n",i,lop->op_name,lop->op_type);
    printf("op_pre : %p\n",lop->op_pre);
    if (lop->op_pre)
      printf("  Data : %s\n",lop->op_pre);
      */
    
    /* First check if the keyword is in the section we are looking for */
    if (IS_OPTION_TYPE_FOR_SECTION (lop->op_type,sec_id) &&

        /* Then make sure keyword is not a child ... */
        (keyword_check_not_child(lops,lop->op_name)) &&
        
         /* Finally make sure keyword is not words "image" & "other" */
        (strcasecmp(lop->op_name,"image") != 0 &&
         strcasecmp(lop->op_name,"other") != 0)) {

      ptr = malloc(sizeof(POSSIBLE_KEYWORDS));

      new_word = calloc (strlen(lop->op_name)+1,sizeof(char));
      strcpy(new_word,lop->op_name);

      ptr->name = new_word;

      ptr->childs = get_childs_for_keyword(lops,ptr->name,sec_id);

      if (lop->op_post) /* Is the word supposed to be "grouped" with childs ? */
        ptr->group = 1; /* Yes, group. Example "MAP-DRIVE & TO" */
      else
        ptr->group = 0; /* no */

/*      printf("%20s : Grouping is %d, Childs : %p\n",ptr->name,ptr->group,ptr->childs);*/
      
      llist_insert (ret_list,ptr);
    }
  }
  /*
   * Apply enable info to each keyword in the list
   */

  keyword_check_if_enabled (lops, section_glob,
                            section_curr,ret_list,sec_id);

  free_options(lops);
  free_option_pre_req(section_glob);
  free_option_pre_req(section_curr);

  return ret_list;
}


/*
 * Frees all keywords
 */
void free_keywords_for(LLIST *keywords)
{
  POSSIBLE_KEYWORDS *ptr;
  int i;

  if (keywords == NULL)
    return;

  i = llist_get_size(keywords);

  while (i-->0) {
    ptr = llist_pop_min(keywords);
    free (ptr->name);
    /* Free all childs with recurse */
    if (ptr->childs != NULL)
      free_keywords_for(ptr->childs);
    free (ptr);
  }
  llist_free(keywords);
}

/*
 * Internal function to print the id's in a LLIST
 */
void print_ids(LLIST *section)
{
  int i,max;
  int *value;

  max = llist_get_size (section);
  printf("There are %d nodes in the list\n",max);
  for (i=0;i<max;i++) {
    value = llist_get_node(section,i);
    printf("%d. %d\n",i,*value);
  }
}
