/* A simple linked list
 * Author : Per Winkvist - (Per.Winkvist@UK.COM)
 * Version 1998-12-01
 */

/* TODO :
 * - Add insert sorted
 * - Add Remove_Node(LLIST *ll, int node);
 * - Add Insert_at_node (LLIST *ll,void *data, int node);
 */

#include "llist.h"
#include "memwatch.h"
#include <signal.h>

/* Returns a new alloced llist */
LLIST *llist_new (void)
{
  LLIST *newlist;
  
  if ((newlist = malloc(sizeof(LLIST))) == NULL) {
    printf ("Couldn't callocate memory! %s:%d\n",__FILE__,__LINE__);
    exit (EXIT_FAILURE);
  }
  
  newlist->head = NULL;
  newlist->count = 0;

  return newlist;
}


void llist_free(LLIST *ll)
{
  if (ll->count != 0) {
    printf ("LList counter is %d !!!\n",ll->count);
    printf ("The llist node counter was not 0\n");
    printf ("Make sure node datapointers are cleared\n");
    printf ("Exiting now...\n");
    /* I consider this serious enough to exit! */
    exit (EXIT_FAILURE);
  }
  free (ll); /* Finally free the listpointer */
}

LLIST_NODE *ll_node_new (void)
{
  LLIST_NODE *ll;
  
  if ((ll = (LLIST_NODE*) malloc(sizeof(LLIST_NODE))) == NULL) {
    printf ("Couldn't callocate memory! %s:%d\n",__FILE__,__LINE__);
    exit (EXIT_FAILURE);
  }
  ll->dataptr = NULL;
  ll->next = NULL;
  
  return(ll);
}

/* Inserts *data at the end of the list */
/* It returns the actual position in the list */
/* This routine should be called llist_append :) */
int llist_insert (LLIST *ll, void *data)
{
  LLIST_NODE *new_node;
  LLIST_NODE *ptr;

  new_node = ll_node_new();
  new_node->dataptr = data;

  ll->count++;

  if (ll->head == NULL) {
    ll->head = new_node; /* Set the first entry */
    return ll->count-1; /* Return position for data */
  }

  ptr = ll->head;
  
  while (ptr->next != NULL) {
    ptr = ptr->next; /* move on */
  }
  
  new_node->next = ptr->next;
  ptr->next = new_node;

  return ll->count-1; /* Return position for data */
}

/* Inserts *data at the position 'position' starting
 * from 0.
 */
int llist_insert_at(LLIST *ll,void *data,int position)
{
  LLIST_NODE *new_node;
  LLIST_NODE *ptr;
  int count=position;

  if (position > ll->count || position < 0) {
    printf("You tried to insert a element into the list\n");
    printf("at pos %d, but there are only %d elements in it!\n",position,ll->count);
    printf("File : %s at line %d  ..... leaving\n",__FILE__,__LINE__);
    exit(EXIT_FAILURE);
  }
  
  new_node = ll_node_new();
  new_node->dataptr = data;

  ll->count++;

  if (ll->head == NULL && position == 0) { /* The list is empty */
    ll->head = new_node;
    return 0; /* return the position */
  }
  else if (position == 0) { /* Insert at the head of the list */
    new_node->next = ll->head;
    ll->head = new_node;
    return 0;
  }

  ptr = ll->head;
  while (--position > 0) /* Find the right position */
    ptr = ptr->next;

  new_node->next = ptr->next;
  ptr->next = new_node;

  return count;
}

/* return the first node in the list */
void *llist_get_min(LLIST *ll)
{
  if (ll == NULL || ll->head == NULL || ll->count < 1) {
    printf ("Error in llist_get_min : ll or ll->head was null!\n");
    printf ("The value of the node counter is %d\n",ll->count);
    exit (EXIT_FAILURE); /* This won't be reached if ll == null) */
  } /*              Since it will seg.fault above on "ll->count" */
  return (void *) ll->head->dataptr;
}

/* pop the lowest in the queue */
void *llist_pop_min(LLIST *ll)
{
  LLIST_NODE *ptr;
  void *dataptr;
  
  llist_get_min(ll); /* Will exit if ll == null or ll->head == null */

  dataptr = (void *) ll->head->dataptr;
  ptr = ll->head; /* get the node to pop */

  ll->head = ll->head->next; /* make ll->head point to next */

  free (ptr); /* Remove the node from memory */
  ll->count--; /* Decrease ref. counter */
  return dataptr; /* return the nodes data pointer */
}

int llist_get_size (LLIST *ll)
{
  if (ll == NULL)
    return 0;
  else
    return ll->count;
}

void *llist_get_node (LLIST *ll, int node)
{
  LLIST_NODE *ptr;

  if (llist_get_size(ll) <= node) {
    printf("You tried to access a node that was outof reach.\n");
    printf("Total nodes = %d. Accessed node = %d\n",llist_get_size(ll),node);
    raise(SIGSEGV);
    exit(EXIT_FAILURE);
  }
  
  ptr = ll->head;

  while (ptr != NULL && node > 0) {
    ptr = ptr->next;
    node--;
  }

  if (ptr != NULL)
    return ptr->dataptr;
  else
    return NULL;
}

void *llist_pop_node (LLIST *ll, int node)
{
  LLIST_NODE *ptr,*del_ptr;
  void *data_ptr;
  int i;

  if (node == 0)
    return llist_pop_min(ll);

  data_ptr = llist_get_node(ll,node); /* get node's data, handle error */

  ptr = ll->head;

  for (i=0;i<node-1;i++) /* go to the node before the one to pop */
    ptr = ptr->next; /* move on */

  /* ptr should now hold node before the one to pop */
  del_ptr = ptr->next; /* Let del_ptr point to the one we wanna delete */
  ptr->next = ptr->next->next; /* Remove the node from the chain */
  
  free (del_ptr); /* Remove the node from memory */
  ll->count--; /* Decrease ref. counter */
  return data_ptr; /* return the nodes data pointer */
}

void llist_for_each (LLIST *ll, LL_FUNC_PTR func)
{

  LLIST_NODE *ptr;

  if (llist_get_size(ll) <= 0)
    printf("Size is <= 0\n");

  llist_get_min (ll);
  ptr = ll->head;

  while (ptr != NULL) {

    (*func) ( ptr->dataptr); /* Call userfunc with ptr to nodedata */

    ptr = ptr->next;
  }

}

/* Adds second list to first one. Returns first list */
LLIST *llist_merge_lists (LLIST *first, LLIST *second)
{
  void *ptr;
  int i,size;

  if (first == NULL) /* First is empty */
    return second;
  if (second == NULL) /* Second is empty */
    return first;
  if (first == second) /* They are the same */
    return first;
  
  
  size = llist_get_size(second);

  for (i=0;i<size;i++) {
    ptr = llist_pop_min(second); /* get the top node */
    llist_insert(first,ptr); /* Insert it in the new list */
  }
  llist_free(second);

  return first; /* And return the modified list (first) */
}

